#include "collide.h"



volatile int cd_error;



static __inline cd_real cd_AXB (m3_vec *A, m3_vec *B)
{
 return A->x*B->y - A->y*B->x;
}

static cd_bool cd_vtx_eq (m3_vtx *p, m3_vtx *q)
{
 return ((p->x == q->x) && (p->y == q->y) && (p->z == q->z));
}



/*
 * search for support vertex
 */
static m3_vtx search_support_vertex (int n, m3_vtx *v, m3_vtx p, m3_vec S)
{
 int i;
 cd_real max, dot;
 m3_vtx newp = p;

 max = m3_dot_product(p.x, p.y, p.z, S.x, S.y, S.z);

 for (i=0; i<n; i++) {
     dot = m3_dot_product(v[i].x, v[i].y, v[i].z, S.x, S.y, S.z);
     if (dot>max) {
        max = dot;
        newp = v[i];
     }
 }

 return newp;
}



/*
 * find half plane
 */
static cd_bool find_half_plane (m3_vec SetR[], int k, m3_vec *w, m3_vec newR)
{
 int i, index;
 m3_rmat M;
 m3_vec ra, rb, T;
 cd_real x, y, z;
 double d;
 cd_bool sw;

 SetR[k] = newR;

 if (m3_dot_product(w->x, w->y, w->z, newR.x, newR.y, newR.z) >= 0.0) {
    return CD_TRUE;
 }
 for (sw=CD_TRUE, i=0;i<k;i++) {
     if (m3_dot_product(w->x, w->y, w->z, SetR[i].x, SetR[i].y, SetR[i].z)!=0.0) {
        sw = CD_FALSE;
        break;
     }
 }
 if (sw) {
    w->x = -w->x;
    w->y = -w->y;
    w->z = -w->z;
    goto finish;
 }

 x = newR.x;
 y = newR.y;
 z = newR.z;
 d = 1.0 /sqrt(SQR(x)+SQR(y));
 /*
              x*z/d    y*z/d   -(x+y)/d 
         M =  -y/d     x/d     0            d = (x+y)
              x        y       z          
 */
#ifndef OPENGL_MATRIX
 M[0][0] = x*z *d;
 M[0][1] = y*z *d;
 M[0][2] = -(SQR(x)+SQR(y)) *d;
 M[1][0] = -y *d;
 M[1][1] = x *d;
 M[1][2] = 0;
 M[2][0] = x;
 M[2][1] = y;
 M[2][2] = z;
#else
 /* the OpenGL spec says the matrix be given in column major order */
 M[0][0] = x*z *d;
 M[1][0] = y*z *d;
 M[2][0] = -(SQR(x)+SQR(y)) *d;
 M[0][1] = -y *d;
 M[1][1] = x *d;
 M[2][1] = 0;
 M[0][2] = x;
 M[1][2] = y;
 M[2][2] = z;
#endif

 m3_apply_rmat(&M, SetR[0], &ra);
 m3_apply_rmat(&M, SetR[1], &rb);
 index = 0;

 if ((cd_AXB(&ra, &rb)==0.0) && ((ra.x!=rb.x)||(ra.y!=rb.y))) {
    /* the vectors are opposed */
    int sign1=0, sign2=0;
    for (i=2; i<k; i++) {
        cd_real t;
        m3_apply_rmat(&M, SetR[i], &T);
        t = cd_AXB(&ra, &T);
        if (t>0.0) {
           sign1 = 1;
        }
        if (t<0.0) {
           sign2 = 1;
        }
    }
    if (sign1 && sign2) {
       return CD_FALSE;
    } else {
       if (sign1) {
          index = 1;
       }
    }
 } else {
    /* the vectors are not opposed */
    if (cd_AXB(&ra, &rb) > 0.0) {
       m3_vec t = ra;
       ra = rb;
       rb = t;
       index = 1;
    }
    /* the vectors are in clock-wise order */
    for (i=2; i<k; i++) {
        cd_real t1, t2;
        m3_apply_rmat(&M, SetR[i], &T);
        t1 = cd_AXB(&ra, &T);
        t2 = cd_AXB(&T, &rb);
        if ((t1<=0.0)&&(t2<=0.0)) {
        } else {
           if ((t1>0.0)&&(t2>0.0)) {
              return CD_FALSE;
           } else {
              if (t1>0.0) {
                 ra = T;
                 index = i;
              } else {
                 if (t2>0.0) {
                    rb = T;
                 }
              }
           }
        }
    }
 }

 m3_cross_product(SetR[index].x, SetR[index].y, SetR[index].z,
                  newR.x, newR.y, newR.z,
                  &w->x, &w->y, &w->z);
 m3_normalize_vector(&w->x, &w->y, &w->z);

finish:
#ifdef CD_PARANOIA
  for (i=0;i<=k;i++) {
      if (m3_dot_product(w->x, w->y, w->z, SetR[i].x, SetR[i].y, SetR[i].z) < -CD_EPSILON) {
         cd_error = CD_HPERR;
         break;
      }
  }
#endif

 return CD_TRUE;
} 



/*
 * collision detection (Chung Tat Leung)
 */
cd_bool separating_vector (int Pn, m3_vtx *Pv,
                           int Qn, m3_vtx *Qv,
                           m3_xmat *Xp, m3_xmat *Xq,
                           m3_rmat *invRp, m3_rmat *invRq)
{
 int i, k, kk;
 cd_real dp;
 cd_bool sw;
 m3_vec S, S1, S2, w, r_[64];
 m3_vtx p,q, pp,qq, saves[64][2];

 cd_error = CD_OK;

 S.x = Xq->t[0] - Xp->t[0];
 S.y = Xq->t[1] - Xp->t[1];
 S.z = Xq->t[2] - Xp->t[2];

 /* Since the polytopes are defined in origin,
  * in case of same translation, we assume the polytopes are colliding.
  */
 if (m3_vector_square(S.x, S.y, S.z)==0.0) {
    return CD_TRUE;
 }
 m3_normalize_vector(&S.x, &S.y, &S.z);

 p = Pv[0];
 q = Qv[0];

 k = kk = 0;

 do {
#ifdef CD_PARANOIA
     if (k>63) {
        cd_error = CD_CYCLE;
     }
#endif
     if (cd_error!=CD_OK) {
        return CD_FALSE;
     }

     m3_apply_rmat(invRp, S, &S1);
     m3_apply_rmat(invRq, S, &S2);
     S2.x = -S2.x;
     S2.y = -S2.y;
     S2.z = -S2.z;

     p = search_support_vertex(Pn, Pv, p, S1);
     q = search_support_vertex(Qn, Qv, q, S2);

     m3_apply_xmat(Xp, p, &pp);
     m3_apply_xmat(Xq, q, &qq);
     r_[k].x = qq.x - pp.x;
     r_[k].y = qq.y - pp.y;
     r_[k].z = qq.z - pp.z;
     m3_normalize_vector(&r_[k].x, &r_[k].y, &r_[k].z);

     dp = m3_dot_product(S.x, S.y, S.z, r_[k].x, r_[k].y, r_[k].z);

     if (dp >= -CD_EPSILON) {
        return CD_FALSE;
     }

     for (sw=CD_FALSE, i=0; i<kk; i++) {
         if (cd_vtx_eq(&p, &saves[i][0]) && cd_vtx_eq(&q, &saves[i][1])) {
            sw = CD_TRUE;
            break;
         }
     }

     if (sw) {
        S = w;
     } else {
        if (k==1) {
           w.x = r_[0].x + r_[1].x;
           w.y = r_[0].y + r_[1].y;
           w.z = r_[0].z + r_[1].z;
           /* w must be the bisector of r_0 and r_1.
            * if (r_0 == -r_1) we manually pick a vector
            */
           if (m3_vector_square(w.x, w.y, w.z)==0.0) {
              if (r_[0].x==0) w.x = 1;
              if (r_[0].y==0) w.y = 1;
              if (r_[0].z==0) w.z = 1;
              if (m3_vector_square(w.x, w.y, w.z)==0.0) {
                 w.x = r_[0].x;
                 w.y = r_[1].y;
                 w.z = 0;
              }
           }
           m3_normalize_vector(&w.x, &w.y, &w.z);
        }
        if ((k>1) && (!find_half_plane(r_, k, &w, r_[k]))) {
           return CD_TRUE;
        }
        saves[kk][0] = p;
        saves[kk][1] = q;
        kk++;

        S.x -= 2*dp*r_[k].x;
        S.y -= 2*dp*r_[k].y;
        S.z -= 2*dp*r_[k].z;
     }

     k++;
 } while (CD_TRUE);
}
