#ifndef MATH3D_H_included
#define MATH3D_H_included

#include <math.h>

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#define SQR(x) ((x)*(x))

/*
 * 3D vertex
 */
typedef struct {
        cd_real x, y, z;
} m3_vtx;

/*
 * 3D vector
 */
typedef struct {
        cd_real x, y, z;
} m3_vec;

/*
 * rotation matrix
 */
typedef cd_real m3_rmat[3][3];

/*
 * translation matrix
 */
typedef cd_real m3_tmat[3];

/*
 * transformation matrix
 */
typedef struct {
        m3_rmat r;
        m3_tmat t;
} m3_xmat;

/*
 * return the square length of a vector
 */
cd_real m3_vector_square (cd_real x, cd_real y, cd_real z);

/*
 * normalize a vector (no test for ZERO)
 */
void m3_normalize_vector (cd_real *x, cd_real *y, cd_real *z);

/*
 * dot product of two vectors
 */
cd_real m3_dot_product (cd_real x1, cd_real y1, cd_real z1,
                        cd_real x2, cd_real y2, cd_real z2);

/*
 * cross product of two vectors
 */
void m3_cross_product (cd_real x1, cd_real y1, cd_real z1,
                       cd_real x2, cd_real y2, cd_real z2,
                       cd_real *xout, cd_real *yout, cd_real *zout);

/*
 * transpose a rotation matrix
 */
void m3_transpose_rmat (m3_rmat *tm, m3_rmat *m);

/*
 * apply a rotation matrix to a vector
 */
void m3_apply_rmat (m3_rmat *m, m3_vec v, m3_vec *w);

/*
 * apply a transformation matrix to a point
 */
void m3_apply_xmat (m3_xmat *m, m3_vtx p, m3_vtx *q);

/*
 * build a transformation matrix
 */
void m3_get_xmat (m3_xmat *m,
                  cd_real xrot, cd_real yrot, cd_real zrot,
                  cd_real x, cd_real y, cd_real z);

#endif
